/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File keyb_int.c - Keyboard interrupt handler for DOS 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * History :
 * Original Author : H. Hanemaayer (hhanemaa@cs.ruu.nl) 
 * (Keyboard I/O based on showkey.c from kbd-0.84 package)
 * DOS port by Mihai Cartoaje ( moisemih@ift.ulaval.ca ).
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <dpmi.h>
#include <go32.h>
#include <pc.h>
#include <sys/nearptr.h>

#include "keyb_int.h"
#include "intdef.h"
#include "../util.h"


#define keybuf_size 200


static __KeyboardHandler __KeyboardEventHandler;

static int KeyboardInitialised = -1;

static int extended = 0;
static int LastKey = 0;
static unsigned char TheCode;

static unsigned char keybuf[keybuf_size]; //buffer klaves
int keys;  //pocet dat v bufferu


/* PC extended keyboard codes prefixed by 0xe0 */
const unsigned char kb_prefix_scancode[128]={
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x60, 0x61, 0x00, 0x00, //PadEnter,RtCtrl
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, //PrtScr
  0x00, 0x00, 0x00, 0x00, 0x00, 0x62, 0x78, 0x63, //?,PrtScr,RtAlt
  0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x77, 0x65, 0x66, 
  0x67, 0x68, 0x00, 0x69, 0x00, 0x6a, 0x00, 0x6b, 
  0x6c, 0x6d, 0x6e, 0x6f, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x70, 0x71, 0x72, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
};

_go32_dpmi_seginfo wrapper, oldHandler, newHandler;

/* Must do our own interrupt handling. Luckily, DOS allows this. */

/* Interrupt driver */


static void 
KbdInt ( void ) 
{
   TheCode = inportb ( 0x60 );

//   printf("keycode = %x, extended = %d\n", TheCode, extended);

   /* Deal with extend keycodes */
   if(extended==1) {
      TheCode = (TheCode & 0x80) | kb_prefix_scancode[TheCode & 0x7f];
      extended = 0;
   }
   else if(extended==2) {
      /* Only handle 0xe0 codes for now */
      TheCode = SCANCODE_A;
      extended = 0;
   }
   else if(TheCode>=0xe0) {
     if(TheCode<=0xe1) extended=TheCode-(0xe0-1);
   }

   /* Insert keystroke into a queue awaiting processing */
   if (!extended && (TheCode != LastKey)) {
      if(keys<keybuf_size) keybuf[keys++]=TheCode;
   }
   outportb(0x20,0x20);    /* Acknowledge the interrupt */
   ENABLE();
}
END_OF_FUNCTION(KbdInt);


int KeyboardInit(__KeyboardHandler Handler)
{
  int error;

  /* Install keyboard handler. */
  __KeyboardEventHandler = Handler;

  if (KeyboardInitialised == -1) {
	
    KeyboardInitialised = 1;

	  wrapper.pm_offset=(int)KbdInt;
	  wrapper.pm_selector=_go32_my_cs();
	  _go32_dpmi_allocate_iret_wrapper(&wrapper);
	  newHandler.pm_offset=wrapper.pm_offset;
	  newHandler.pm_selector=wrapper.pm_selector;

	  LOCK_VARIABLE(wrapper);
	  LOCK_VARIABLE(oldHandler);
	  LOCK_VARIABLE(newHandler);
	  LOCK_VARIABLE(kb_prefix_scancode);
	  LOCK_VARIABLE(extended);
	  LOCK_VARIABLE(LastKey);
	  LOCK_VARIABLE(TheCode);
	  LOCK_FUNCTION(KbdInt);

	  error=_go32_dpmi_get_protected_mode_interrupt_vector(KEYBD_INT, &oldHandler);
	  error+=2*_go32_dpmi_set_protected_mode_interrupt_vector(KEYBD_INT, &newHandler);
    if(error) simcoupe_mesg(SIMCOUPE_FATAL,"cannot install keyboard handler (error %i)",error);

//    simcoupe_mesg(SIMCOUPE_INFO,"old keyboard handler=%p",oldHandler);
  }
  return 0;
}


void KeyboardClose(void)
{
  if(KeyboardInitialised<0) return;

//  simcoupe_mesg(SIMCOUPE_INFO,"restoring old keyboard handler (%p)",oldHandler);
  _go32_dpmi_set_protected_mode_interrupt_vector(KEYBD_INT,&oldHandler);

//  simcoupe_mesg(SIMCOUPE_INFO,"releasing keyboard interrupt wrapper=%p",wrapper);
  _go32_dpmi_free_iret_wrapper(&wrapper);

  KeyboardInitialised=-1;
}


void ProcessKeyBuf(void) {
  int i;
  for(i=0;i<keys;i++) {
    int keycode=keybuf[i];
    __KeyboardEventHandler(keycode&0x7f, (keycode&0x80) ? KEY_EVENTRELEASE : KEY_EVENTPRESS);
  }
  keys=0;
}