/*
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File audio.cpp - audio emulation code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <string.h>

#include "resource.h"
#include "audio.h"
#include "driver.h"
#include "allegro.h"

extern "C" {
#include "ay8910.h"
#include "saa1099.h"
#include "util.h"
#include "samio.h"
}


#if MAX_8910!=1
#error MAX_8910 must be #defined to 1 (file "ay8910.h")
#endif


#if MAX_SAA1099!=1
#error MAX_SAA1099 must be #defined to 1 (file "saa1099.h")
#endif


#define AYclock 1773400    //according to "128k.mat" (CPU clock/2)


BEGIN_DIGI_DRIVER_LIST
  DIGI_DRIVER_AUDIODRIVE
  DIGI_DRIVER_SOUNDSCAPE
  DIGI_DRIVER_WINSOUNDSYS
  DIGI_DRIVER_SB
END_DIGI_DRIVER_LIST

BEGIN_MIDI_DRIVER_LIST
END_MIDI_DRIVER_LIST


//---------------------------------------------------------------------------


static INT16 *buf[4];  //sound buffers for three AY channels + beeper

static void (*sh_callback)(int param,INT16 **buffer,int length);

static int writepos;   //write pos in sample units


const int sample_rate=45454;
const int audio_bits=16;
const int stereo=1;
const int latency=3;

static int bprbegin;
static int bprpos;


//allegro stuff
static SAMPLE *sample;
static int voice;


#define rate50 (sample_rate/50)  //pocet samplu za 1/50 sekundy


//---------------------------------------------------------------------------


int stream_init_multi(
  int channels,const char **names,const int *mixing_levels,
  int sample_rate, int param,
  void (*callback)(int param,INT16 **buffer,int length)) {

  //channels=3 (always)
  //names is ignored
  //sample_rate=Machine->sample_rate
  //param=chip=0 (always)
  //mixing_levels[3] - ignore???
  sh_callback=callback;
}


void stream_update(int channel,int min_interval) {
  return;
}


extern "C" {

int cpu_get_pc(void) {return 0;}

const char* sound_name(int) {return "(sound_name)";}

void mixer_set_volume(int,int) {}

}


//---------------------------------------------------------------------------


int AudioReinit(void) {
//
//reinitialize sound hardware (called when changing machine type)
//
  if(!cfg.audio) return 0;

  //global variable Machine is required during initialization
  //by emptyhead authors of Mame
  SMachine machine;
  machine.sample_rate=sample_rate;
  Machine=&machine;

  struct MachineSound msound;

  if(cfg.zxmode!=EM_Sam) {

    //AY8910 mode
    saa1099_sh_stop();

    struct AY8910interface sndif;

    msound.sound_interface=&sndif;

    memset(&sndif,0,sizeof(sndif));
    sndif.num=1;
    sndif.baseclock=AYclock;

    sh_callback=0;
    return AY8910_sh_start(&msound);

  } else {

    //SAA1099 mode
    struct SAA1099_interface sndif;

    msound.sound_interface=&sndif;

    memset(&sndif,0,sizeof(sndif));
    sndif.numchips=1;

    sh_callback=0;
    return saa1099_sh_start(&msound);
  }
}


//---------------------------------------------------------------------------


void AudioClose(void) {
//
//close audio emulation
//
  saa1099_sh_stop();

  for(int i=0;i<4;i++) {delete[] buf[i];buf[i]=0;}
  sh_callback=0;

  //allegro stuff
  if(sample) {
    voice_stop(voice);
    deallocate_voice(voice);
    destroy_sample(sample);
    sample=0;
  }

  writepos=0;

  bprbegin=bprpos=0;
}


//---------------------------------------------------------------------------


int AudioInit(void) {
//
//initialize audio emulation
//
//return: 0=ok, else error
//
  static int first_time=1;
  if(first_time) {
    OnQuit(AudioClose,"AudioClose()");
    first_time=0;
  }

  if(!cfg.audio) return 0;

  /* initialize allegro for sound output */
	reserve_voices(1,0);
  install_sound(DIGI_AUTODETECT,MIDI_NONE,0);

  return AudioReinit();
}


//---------------------------------------------------------------------------


void AudioUpdate(void) {
//
//called often to fill buffers
//
  if(!cfg.audio) return;

  //at first call of this function go and create all required buffers
  static int first=1;
  if(first) {
    first=0;

    //allocate 4 channel buffers (3*AY + beeper)
    for(int i=0;i<4;i++) buf[i]=new INT16[rate50*latency];

    //empty begin of beeper's buffer
    bprpos=bprbegin=rate50/2; //latency
    for(int i=0;i<rate50/2;i++) buf[3][i]=0x8000;

	  sample=create_sample(audio_bits,stereo,sample_rate,rate50*latency);
    if(!sample) return;

    voice=allocate_voice(sample);
    voice_set_playmode(voice,PLAYMODE_LOOP);

    //clear buffer
    INT16 *buf=(INT16*)sample->data;
    int p=(rate50*latency<<stereo)-1;
    while(p) buf[p--]=(INT16)0x8000;

    voice_start(voice);
  }

  if(!sample) return;

  //try to keep buffer full
  //how many samples must we update?
  INT16 *w=(INT16*)sample->data+(writepos<<stereo);
  int ahead=voice_get_position(voice)-writepos;
  int size1,size2;

  if(!ahead) return; //not playing?

  if(ahead>0) {
    //one continuous buffer
    size1=ahead;
    writepos+=ahead;
  } else {
    //buffer is interrupted into two parts
    ahead+=rate50*latency;
    size1=rate50*latency-writepos;
    writepos=ahead-size1;
  }

  //fill channel buffers
  INT16 *buffer[3]={buf[0],buf[1],buf[2]};
  sh_callback(0,buffer,ahead);

  //fill-up beeper buffer (if required)
//  printf("bprpos=%i, left %i\n",bprpos,bprpos-ahead);
  if(bprpos<ahead) {
    //repeat last byte (better than nothing)
    INT16 value=buf[3][bprpos-1];
    for(;bprpos<ahead;bprpos++) buf[3][bprpos]=value;
  }

  if(GetCfg(zxmode)!=0) {

   //AY8910

   //combine all three AY channels together
   if(!stereo) {

    //mono
    int i=0;
    for(;i<size1;i++) {
      *w=((long)buf[0][i]+buf[1][i]+buf[2][i]+buf[3][i]);
      w++;
    }
    //second half
    w=(INT16*)sample->data;
    for(;i<ahead;i++) {
      *w=((long)buf[0][i]+buf[1][i]+buf[2][i]+buf[3][i]);
      w++;
    }

   } else {

    //stereo
    int i=0;
    for(;i<size1;i++) {
      w[0]=w[1]=((long)buf[0][i]+buf[1][i]+buf[2][i]+buf[3][i]);
      w+=2;
    }
    //second half
    w=(INT16*)sample->data;
    for(;i<ahead;i++) {
      w[0]=w[1]=((long)buf[0][i]+buf[1][i]+buf[2][i]+buf[3][i]);
      w+=2;
    }

   }
  } else {

    //SAA1099
    if(!stereo) {
/*
      //mono
      int i=0;
      for(;i<size1;i++) {
        *w=((long)buf[0][i]+buf[1][i]+buf[2][i]);
        w++;
      }
      //second half
      w=(INT16*)sample->data;
      for(;i<ahead;i++) {
        *w=((long)buf[0][i]+buf[1][i]+buf[2][i]);
        w++;
      }
*/
    } else {

      //stereo
      int i=0;
      for(;i<size1;i++) {
        w[0]=buf[0][i]+(short)32768+buf[3][i];
        w[1]=buf[1][i]+(short)32768+buf[3][i];
        w+=2;
      }
      //second half
      w=(INT16*)sample->data;
      for(;i<ahead;i++) {
        w[0]=buf[0][i]+(short)32768+buf[3][i];
        w[1]=buf[1][i]+(short)32768+buf[3][i];
        w+=2;
      }
    }
  }

  //prebytecna data speakeru presune na zacatek bufferu
  if(bprpos>=ahead) {
    for(int i=bprpos-ahead-1;i>=0;i--) buf[3][i]=buf[3][ahead+i];
    bprbegin=bprpos-ahead;
    if(bprbegin>rate50) bprbegin--;
    //nesmi to rust do nekonecna - aspon jeden rate50 blok musi zustat volny
    if(bprbegin>(latency-1)*rate50) bprbegin=(latency-1)*rate50;
  } else {
    bprbegin=0;
  }
  bprpos=bprbegin;
}


//---------------------------------------------------------------------------


void BeeperNextLine(void) {
//
//volano na zacatku nove linky obrazu
//
//beeper values: 12bit unsigned --> 0/16384
//
  extern int totallines;
  extern int LineNo;

  if(!cfg.audio || !buf[3]) return;

  int pos=bprbegin+LineNo*rate50/totallines;

  int value=beeperbit*16384;
  for(;bprpos<pos;bprpos++) buf[3][bprpos]=value;
}


//---------------------------------------------------------------------------


void AudioMute(void) {
//
//mute sound
//
  //clear buffer
  if(sample && sample->data) {
    INT16 *buf=(INT16*)sample->data;
    int p=rate50*latency<<stereo;
    while(p) buf[--p]=(INT16)0x8000;
  }
}

