/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File video.h - video emulation code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <string.h>

#include "allegro.h"

#include "z80.h"
#include "dtbl.h"
#include "video.h"
#include "util.h"
#include "resource.h"
#include "mem.h"


#define OSD_DELAY 125  //time in frames


/* These macros map a byte in the attribute area to a vga color index */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)


#define draw1(y) drawline((y),320)
#define draw2(y) drawline((y),640)


#define log(text) simcoupe_mesg(SIMCOUPE_INFO,text)


BEGIN_GFX_DRIVER_LIST
  GFX_DRIVER_VBEAF
  GFX_DRIVER_VESA3
  GFX_DRIVER_VESA2L
  GFX_DRIVER_VESA2B
  GFX_DRIVER_VESA1
  GFX_DRIVER_VGA
END_GFX_DRIVER_LIST


BEGIN_COLOR_DEPTH_LIST
  COLOR_DEPTH_8
END_COLOR_DEPTH_LIST


//---------------------------------------------------------------------------


#define SizeW1x1 320

#define SizeW2x2 640
#define SizeH2x2 480

#define BorderW_1x1 ((SizeW1x1-256)/2)
#define BorderW_2x2 ((SizeW2x2-512)/2)

#define BorderH_2x2 ((SizeH2x2-384)/2)  //vyska borderu


//---------------------------------------------------------------------------


static struct RGB palette[128];
static int scalefactor;   //1 or 2

static int osd_delay;     //for osd only
static const char *osd_text;  //osd text
static int rescaledown;   //1,2,4=wanted rescale to 1x1 for specified mode
static int mode3begin;    //1=mode 3 has begun in the current frame

unsigned long pixels[128];//value representing a color (in DOS always pixels[i]=i)

//port 0xff handling
static int attr_tbegin;
static uns8* SamAttrPtr;
extern int LineCycleCounter;

extern int totallines;

//the following array must cover whole screen, so I declared it [312] for sure
char LineNeedsUpdating[312]; /* Determines whether lines need to be redrawn */


uns8 TmpBuff[640];        //single line buffer


/* General level function pointers */
void (*borderline_update)(void);
void (*line_update)(int);
void (*line_update_mode_1)(int);
void (*line_update_mode_2)(int);
void (*line_update_mode_3)(int);
void (*line_update_mode_4)(int);


int b_lines_drawn;    //border lines drawn, usually (240-192)/2=24
extern int toplines;  //top lines between frame interrupt and paper


static int BorderY = 0;
static uns8 *SamMemoryPtr;   /* Pointer to the SAM screen memory */


extern int flash_phase;


//---------------------------------------------------------------------------


static inline void drawline(const int y,int len) {
//
//draws a line using far memory transfers
//
  unsigned long ofs=bmp_write_line(screen,y);
  int seg=screen->seg;
  len>>=2;

  asm __volatile__ ("
    push %%es
    cld
    movw %%bx, %%es
    rep
    movsl
    pop %%es
    "
    : 
    : "b" (seg), "S" (TmpBuff), "D" (ofs), "c" (len)
    : "cc"
  );
}


//---------------------------------------------------------------------------


static void initcolormap(void) {
//
//initialize 128-colour palette
//
  int i;
  int red_index, green_index, blue_index;

  static const int RGBvals[8]={0x00,0x09,0x12,0x1b,0x24,0x2d,0x36,0x3f};

  for(i=0;i<128;i++) {
    blue_index  = ((i & 0x01) << 1) | ((i & 0x10) >> 2) | ((i & 0x08) >> 3);
    red_index   = ((i & 0x02)     ) | ((i & 0x20) >> 3) | ((i & 0x08) >> 3);
    green_index = ((i & 0x04) >> 1) | ((i & 0x40) >> 4) | ((i & 0x08) >> 3);

    palette[i].r=RGBvals[red_index];
    palette[i].g=RGBvals[green_index];
    palette[i].b=RGBvals[blue_index];

    pixels[i]=(long)i;
  }
}


//---------------------------------------------------------------------------


void VideoClose(void) {
//
//close video emulation
//
  free((void*)osd_text);
  osd_text=0;

  allegro_exit();
}


//---------------------------------------------------------------------------


void VideoInit(void) {
//
//initialize video emulation
//
  allegro_init();

  flash_phase=!cfg.flash;

  /* Initialize variables used to keep track of line interrupt clut changes */
  ClutChangeInLine=0;
  BorderChangeInLine=0;
  VmprChangeInLine=0;

  initcolormap();

//  VideoSetScale(GetCfg(novesa) ? 1 : 2);
  VideoSetScale(1);
  VideoSetMode(1);

  OnQuit(VideoClose,"VideoClose()");
}


//---------------------------------------------------------------------------


static void update_osd(void) {
//
//update osd info
//
  if(osd_delay) {
    textout(screen,font,(unsigned char*)osd_text,0,0,127);

    if(!--osd_delay) {
      //simply redraw top 8 lines of the screen to remove osd text
      if(BorderChangeInLine<toplines-b_lines_drawn+9) {
        BorderChangeInLine=toplines-b_lines_drawn+9;
      }

      //erase text for sure (needed when scanlines=1)
      if(cfg.scanlines) textout(screen,font,(unsigned char*)osd_text,0,0,0);

      free((void*)osd_text);
      osd_text=0;
    }
/*
  } else {
    char buf[1000];
    Z80_Regs r;
    get_registers(&r);
    //zobrazuje PC a navratovou adresu
    sprintf(buf,"pc=%04X, (sp)=%04X",r.pc,read_word(r.sp));
    textout(screen,font,(unsigned char*)buf,0,0,127);
*/
  }

  //EmuWatch!
  if(cfg.mon && scalefactor>1) {
    char buf[200];
    Z80_Regs r;
    int y=SCREEN_H-48;
    int x=20*8;
    int i;

    get_registers(&r);

#define w(h,l) (r.h*256+r.l)
    sprintf(buf,"PC=%04x SP=%04x I=%2x R=%02x %cI iff2=%cI",
            r.pc,r.sp,r.i,r.r,r.iff1 ? 'E' : 'F',r.iff2 ? 'E' : 'F');
    textout(screen,font,(unsigned char*)buf,x,y,127);
    sprintf(buf,"A=%02x BC=%04x DE=%04x HL=%04x IX=%04x IY=%04x",
            r.a,w(b,c),w(d,e),w(h,l),r.ix,r.iy);
    textout(screen,font,(unsigned char*)buf,x,y+8,127);
    sprintf(buf,"A'%02x BC'%04x DE'%04x HL'%04x F=%02x F'%02x",
            r.a1,w(b1,c1),w(d1,e1),w(h1,l1),r.f,r.f1);
    textout(screen,font,(unsigned char*)buf,x,y+16,127);
    sprintf(buf,"lmpr=%02x hmpr=%02x vmpr=%02x lepr=%02x hepr=%02x "
                "bord=%02x line=%02x",
            lmpr,hmpr,vmpr,lepr,hepr,
            border,line_int);
    textout(screen,font,(unsigned char*)buf,x,y+24,127);

    for(i=0;i<6;i++) {
      sprintf(buf,"%s%04x=%04x",i==3 ? "SP=>" : "    ",
              r.sp+6-2*i,read_word(r.sp+6-2*i));
      textout(screen,font,(unsigned char*)buf,0,y+8*i,127);
    }

#undef w
  }
}


//---------------------------------------------------------------------------


void osd(const char *text) {

  if(osd_delay) {
    //simply redraw top of the screen to remove osd text
    if(BorderChangeInLine<toplines-b_lines_drawn+9) {
      BorderChangeInLine=toplines-b_lines_drawn+9;
    }

    //erase text for sure (needed when scanlines=1)
    if(cfg.scanlines) textout(screen,font,(unsigned char*)osd_text,0,0,0);

    free((void*)osd_text);
  }

  osd_text=strdup(text);
  osd_delay=OSD_DELAY;
}


//---------------------------------------------------------------------------


int read_attr(void) {
//
//input from port 0xff = attribute byte
//
  int tstates=LineCycleCounter-attr_tbegin;
  return tstates<128 ? SamAttrPtr[tstates/2] : 0xff;
}


//---------------------------------------------------------------------------


static void borderline_update_1x1(void) {

  if(cfg.mon && scalefactor>1 && BorderY>100) return;

  /* Update either top or bottom borders if required */
  if(ClutChangeInLine || BorderChangeInLine) {
    //draw horizontal line
    hline(screen,0,BorderY,319,clut[BORD_VAL(border)]);
  }
  BorderY++;
  
  if(BorderY==b_lines_drawn) {
    BorderY=SCREEN_H-b_lines_drawn;
    update_osd();
  }
}


static void line_update_mode1_1x1(int LineNo) {

  int i;
  
//  attr_tbegin=LineCycleCounter;

  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_1x1;

    /* In Mode 1 we have to update SamMemoryPtr each line due to the Spectrum
       screen layout, but we use a lookup table to minimize the cost */
    SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
    SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);

    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_1x1);

    /* The Actual screen line */
    for(i=0;i<32;i++) {
      int attr=*SamAttrPtr++;
      int ink=Foreground(attr);
      int paper=Background(attr);
      int i_pix,p_pix,val;
      
      /* toggle the colours if we're in the inverse part of the FLASH cycle */
      if(flash_phase && (attr & 0x80)) {
        int tmp=ink; ink=paper; paper=tmp;
      }

      i_pix = clut[ink];
      p_pix = clut[paper];

      val=*SamMemoryPtr;
      TmpPtr[0] = (val & 0x80) ? i_pix : p_pix;
      TmpPtr[1] = (val & 0x40) ? i_pix : p_pix;
      TmpPtr[2] = (val & 0x20) ? i_pix : p_pix;
      TmpPtr[3] = (val & 0x10) ? i_pix : p_pix;
      TmpPtr[4] = (val & 0x08) ? i_pix : p_pix;
      TmpPtr[5] = (val & 0x04) ? i_pix : p_pix;
      TmpPtr[6] = (val & 0x02) ? i_pix : p_pix;
      TmpPtr[7] = (val & 0x01) ? i_pix : p_pix;
      SamMemoryPtr++;
      TmpPtr+=8;
    }

    /* Rhs Border */
    memset(TmpBuff+(BorderW_1x1+256), clut[BORD_VAL(border)], BorderW_1x1);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw1(LineNo+b_lines_drawn);
  }
}


static void line_update_mode2_1x1(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr once in a given frame, at LineNo == 0 */
  if(VmprChangeInLine || !LineNo)
    SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_1x1;
    
    /* Lhs Border */
    memset(TmpBuff,clut[BORD_VAL(border)],BorderW_1x1);

    /* The Actual screen line */
    SamAttrPtr=SamMemoryPtr+0x2000;
    for(i=0;i<32;i++) {
      int attr=*SamAttrPtr++; 
      int ink=Foreground(attr);
      int paper=Background(attr);
      int i_pix,p_pix,val;
      
      /* toggle the colours if we're in the inverse part of the FLASH cycle */
      if(flash_phase && (attr & 0x80)) {
        int tmp=ink; ink=paper; paper=tmp;
      }

      i_pix = clut[ink];
      p_pix = clut[paper];
      val=*SamMemoryPtr;
      TmpPtr[0] = (val & 0x80) ? i_pix : p_pix;
      TmpPtr[1] = (val & 0x40) ? i_pix : p_pix;
      TmpPtr[2] = (val & 0x20) ? i_pix : p_pix;
      TmpPtr[3] = (val & 0x10) ? i_pix : p_pix;
      TmpPtr[4] = (val & 0x08) ? i_pix : p_pix;
      TmpPtr[5] = (val & 0x04) ? i_pix : p_pix;
      TmpPtr[6] = (val & 0x02) ? i_pix : p_pix;
      TmpPtr[7] = (val & 0x01) ? i_pix : p_pix;
      SamMemoryPtr++;
      TmpPtr+=8;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_1x1+256), clut[BORD_VAL(border)], BorderW_1x1);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw1(LineNo+b_lines_drawn);

  } else {
    /* Short route */
    SamMemoryPtr+=32;             /* 32 bytes in a mode 2 line */
  }
}


static void line_update_mode3_1x1(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
     at LineNo == 128 (when we cross the page boundry ) */
  if(VmprChangeInLine || !(LineNo&127)) {
    SamMemoryPtr = (LineNo & 0x80) ? 
      realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
      realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
  }
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_1x1;
    int md3col=HMPR_MD3COL;
    uns8 pal[4]={clut[md3col],clut[md3col+2],clut[md3col+1],clut[md3col+3]};

    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_1x1);
    
    /* The Actual screen line */
    for(i=0;i<128;i++) {
      int val=*SamMemoryPtr;
      TmpPtr[0]=pal[(val&0x30)>>4];
      TmpPtr[1]=pal[val&3];
      TmpPtr+=2;
      SamMemoryPtr++;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_1x1+256), clut[BORD_VAL(border)], BorderW_1x1);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw1(LineNo+b_lines_drawn);

  } else {
    /* Short route */
    SamMemoryPtr+=128;            /* 128 bytes in a mode 3 line */
  }
}


static void line_update_mode4_1x1(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
     at LineNo == 128 (when we cross the page boundry ) */
  if(VmprChangeInLine || !(LineNo&127)) {
    SamMemoryPtr = (LineNo & 0x80) ?
      realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
      realMem[ VMPR_SCREEN_PAGE          ] + (LineNo<<7);
  }
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_1x1;
    
    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_1x1);
    
    /* The Actual screen line */
    for(i=0;i<128;i++) {
      int val=*SamMemoryPtr;
      TmpPtr[0]=clut[val>>4];
      TmpPtr[1]=clut[val&0xf];
      SamMemoryPtr++;
      TmpPtr+=2;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_1x1+256), clut[BORD_VAL(border)], BorderW_1x1);
    
    LineNeedsUpdating[LineNo]=0;
    
    draw1(LineNo+b_lines_drawn);
  }
  else {
    /* Short route */
    SamMemoryPtr+=128;            /* 128 bytes in a mode 4 line */
  }
}


//---------------------------------------------------------------------------


static void borderline_update_2x2(void) {

  if(cfg.mon && scalefactor>1 && BorderY>100) return;

  /* Update either top or bottom borders if required */
  if(ClutChangeInLine || BorderChangeInLine) {
    hline(screen,0,BorderY,639,clut[BORD_VAL(border)]);
    if(!cfg.scanlines) hline(screen,0,BorderY+1,639,clut[BORD_VAL(border)]);
  }
  BorderY+=2;
  
  if(BorderY==BorderH_2x2) {
    BorderY=SizeH2x2-BorderH_2x2;
    update_osd();
  }
}


static void line_update_mode1_2x2(int LineNo) {

  int i;
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_2x2;
    
    /* In Mode 1 we have to update SamMemoryPtr each line due to the Spectrum
       screen layout, but we use a lookup table to minimize the cost */
    SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + y_dtbl[LineNo];
    SamAttrPtr   = realMem[VMPR_SCREEN_PAGE] + 6144+((LineNo & 0xf8)<<2);
//    attr_tbegin=LineCycleCounter;

    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_2x2);
    
    /* The Actual screen line */
    for(i=0;i<32;i++) {
      int attr=*SamAttrPtr++;
      int ink=Foreground(attr);
      int paper=Background(attr);
      int i_pix,p_pix,val;

      /* toggle the colours if we're in the inverse part of the FLASH cycle */
      if(flash_phase && (attr&0x80)) {
        int tmp=ink; ink=paper; paper=tmp;
      }
      
      i_pix = clut[ink];
      p_pix = clut[paper];
      val=*SamMemoryPtr;
      TmpPtr[0] = TmpPtr[1] = (val & 0x80) ? i_pix : p_pix;
      TmpPtr[2] = TmpPtr[3] = (val & 0x40) ? i_pix : p_pix;
      TmpPtr[4] = TmpPtr[5] = (val & 0x20) ? i_pix : p_pix;
      TmpPtr[6] = TmpPtr[7] = (val & 0x10) ? i_pix : p_pix;
      TmpPtr[8] = TmpPtr[9] = (val & 0x08) ? i_pix : p_pix;
      TmpPtr[10]= TmpPtr[11]= (val & 0x04) ? i_pix : p_pix;
      TmpPtr[12]= TmpPtr[13]= (val & 0x02) ? i_pix : p_pix;
      TmpPtr[14]= TmpPtr[15]= (val & 0x01) ? i_pix : p_pix;
      SamMemoryPtr++;
      TmpPtr+=16;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_2x2+512), clut[BORD_VAL(border)], BorderW_2x2);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw2((LineNo<<1)+BorderH_2x2);
    if(!cfg.scanlines) draw2(1+(LineNo<<1)+BorderH_2x2);
  }
}


static void line_update_mode2_2x2(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr once in a given frame, at LineNo == 0 */
  if(VmprChangeInLine || !LineNo)
    SamMemoryPtr = realMem[VMPR_SCREEN_PAGE] + (LineNo<<5);
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine || 
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *SamAttrPtr=SamMemoryPtr+0x2000;
    uns8 *TmpPtr=TmpBuff+BorderW_2x2;
    
    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_2x2);
    
    /* The Actual screen line */
    for(i=0;i<32;i++) {
      int attr=*SamAttrPtr++; 
      int ink=Foreground(attr);
      int paper=Background(attr);
      int i_pix,p_pix,val;
      
      /* toggle the colours if we're in the inverse part of the FLASH cycle */
      if(flash_phase && (attr & 0x80)) {
        int tmp=ink; ink=paper; paper=tmp;
      }

      i_pix = clut[ink];
      p_pix = clut[paper];
      val=*SamMemoryPtr;
      TmpPtr[0] = TmpPtr[1] = (val & 0x80) ? i_pix : p_pix;
      TmpPtr[2] = TmpPtr[3] = (val & 0x40) ? i_pix : p_pix;
      TmpPtr[4] = TmpPtr[5] = (val & 0x20) ? i_pix : p_pix;
      TmpPtr[6] = TmpPtr[7] = (val & 0x10) ? i_pix : p_pix;
      TmpPtr[8] = TmpPtr[9] = (val & 0x08) ? i_pix : p_pix;
      TmpPtr[10]= TmpPtr[11]= (val & 0x04) ? i_pix : p_pix;
      TmpPtr[12]= TmpPtr[13]= (val & 0x02) ? i_pix : p_pix;
      TmpPtr[14]= TmpPtr[15]= (val & 0x01) ? i_pix : p_pix;
      SamMemoryPtr++;
      TmpPtr+=16;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_2x2+512), clut[BORD_VAL(border)], BorderW_2x2);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw2((LineNo<<1)+BorderH_2x2);
    if(!cfg.scanlines) draw2(1+(LineNo<<1)+BorderH_2x2);
  } else {
    /* Short route */
    SamMemoryPtr+=32;             /* 32 bytes in a mode 2 line */
  }
}


static void line_update_mode3_2x2(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
     at LineNo == 128 (when we cross the page boundry ) */
  if(VmprChangeInLine || !(LineNo&127)) {
    SamMemoryPtr = (LineNo & 0x80) ? 
      realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
      realMem[VMPR_SCREEN_PAGE]+(LineNo<<7);
  }
  
  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    int md3col=HMPR_MD3COL;
    uns8 pal[4]={clut[md3col],clut[md3col+2],clut[md3col+1],clut[md3col+3]};
    uns8 *TmpPtr=TmpBuff+BorderW_2x2;
    
    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_2x2);
    
    /* The Actual screen line */
    for(i=0;i<128;i++) {
      int val=*SamMemoryPtr;
      TmpPtr[0] = pal[val>>6];
      TmpPtr[1] = pal[(val&0x30)>>4];
      TmpPtr[2] = pal[(val&0x0c)>>2];
      TmpPtr[3] = pal[val&3];
      SamMemoryPtr++;
      TmpPtr+=4;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_2x2+512), clut[BORD_VAL(border)], BorderW_2x2);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw2((LineNo<<1)+BorderH_2x2);
    if(!cfg.scanlines) draw2(1+(LineNo<<1)+BorderH_2x2);
  } else {
    /* Short route */
    SamMemoryPtr+=128;            /* 128 bytes in a mode 3 line */
  }
}


void line_update_mode4_2x2(int LineNo) {

  int i;
  
  /* If there is no flagged vmpr change then we only need to update 
     SamMemoryPtr twice in a given frame, once at LineNo == 0 and once
     at LineNo == 128 (when we cross the page boundry ) */
  if(VmprChangeInLine || !(LineNo&127)) {
    SamMemoryPtr = (LineNo & 0x80) ? 
      realMem[(VMPR_SCREEN_PAGE+1) & 0x1f] + ((LineNo & 0x7f)<<7) :
      realMem[VMPR_SCREEN_PAGE]+(LineNo<<7);
  }

  /* Now we have a long or a short route either we need to update the frame
     buffer (if the memory has been written to, or we have a vmpr/clut change
     to deal with), or we just need to update the read/write pointers */
  if(VmprChangeInLine || ClutChangeInLine ||
    BorderChangeInLine || LineNeedsUpdating[LineNo]) {
    /* Long route */   
    uns8 *TmpPtr=TmpBuff+BorderW_2x2;
    
    /* If we are using a paged vga mode for 2x2 scaling we need to deal with
       the switch over lines in a special way */
    
    /* Lhs Border */
    memset(TmpBuff, clut[BORD_VAL(border)], BorderW_2x2);
    
    /* The Actual screen line */
    for(i=0;i<128;i++) {
      int val=*SamMemoryPtr;
      TmpPtr[0] =
      TmpPtr[1] = clut[val>>4];
      TmpPtr[2] =
      TmpPtr[3] = clut[val&0xf];
      SamMemoryPtr++;
      TmpPtr+=4;
    }
    
    /* Rhs Border */
    memset(TmpBuff+(BorderW_2x2+512), clut[BORD_VAL(border)], BorderW_2x2);
    
    LineNeedsUpdating[LineNo] = False;
    
    draw2((LineNo<<1)+BorderH_2x2);
    if(!cfg.scanlines) draw2(1+(LineNo<<1)+BorderH_2x2);
  } else {
    /* Short route */
    SamMemoryPtr+=128;            /* 128 bytes in a mode 4 line */
  }
}


//---------------------------------------------------------------------------


void FlyBack(void) {
  //reset the border line counter
  BorderY=0;

  //delayed rescale to 1x1
  if(rescaledown) {
    //apply only if mode3begin==0 i.e. mode 3 wasn't used for a while
    if(!mode3begin) {
      VideoSetScale(1);
      VideoSetMode(rescaledown);
    }
    rescaledown=0;
  }

  //decrement mode 3 hold-on timer
  if(mode3begin) mode3begin--;
}


//---------------------------------------------------------------------------


static int set_vga_mode(int scale) {
//
//set scaled VGA mode
//
//return: error flag
//
  int modechanged=0;

  if(!cfg.vesa) {
    //VGA 320x200 failsafe mode
    if(scale!=1 || set_gfx_mode(GFX_VGA,320,200,0,0)) return -1;
    b_lines_drawn=4;
    cfg.autoscale=0;

  } else switch(scale) {

  case 1:
    if(!set_gfx_mode(GFX_AUTODETECT,320,240,0,0)) {
      //320x240
      b_lines_drawn=24;//(240-192)/2

      modechanged=1;
    } else {
      //320x200
      if(set_gfx_mode(GFX_AUTODETECT,320,200,0,0)) return -1;
      b_lines_drawn=4; //(200-192)/2
    }
    break;

  case 2:
    if(set_gfx_mode(GFX_AUTODETECT,640,480,0,0)) return -1;
    b_lines_drawn=24;
    modechanged=1;
    break;
  }

  //must set palette
  set_palette_range(palette,0,127,0);

  if(modechanged) {
    char buf[20];
    sprintf(buf,"%dx%d %s",gfx_driver->w,gfx_driver->h,cfg.autoscale?"auto":"");
    osd(buf);
  }

  return 0;
}


//---------------------------------------------------------------------------


void VideoTextMode(void) {
//
//switch to text mode
//
  set_gfx_mode(GFX_TEXT,80,25,0,0);
  scalefactor=0;
}


//---------------------------------------------------------------------------


int VideoGetScale(void) {
  return scalefactor;
}


//---------------------------------------------------------------------------


void VideoSwitchScale(void) {
  if(!cfg.autoscale) VideoSetScale(scalefactor==1 ? 2 : 1);
}


//---------------------------------------------------------------------------


void VideoSetScale(int value) {

  if(!cfg.vesa) value=1; //VGA can't have scale=2
  if(value==scalefactor) return; //no changes needed

  //2x2
  if(value==2) {

    if(!set_vga_mode(2)) {

      borderline_update  = borderline_update_2x2;
      line_update_mode_1 = line_update_mode1_2x2;
      line_update_mode_2 = line_update_mode2_2x2;
      line_update_mode_3 = line_update_mode3_2x2;
      line_update_mode_4 = line_update_mode4_2x2;
    } else {
      //mode not available
      value=1;
    }
  }

  //1x1
  if(value==1) {
    if(set_vga_mode(1)) {
      //fatal error - no graphics mode available!
      simcoupe_mesg(SIMCOUPE_FATAL,"Cannot find any suitable graphics mode!");
    }

		borderline_update  = borderline_update_1x1;
		line_update_mode_1 = line_update_mode1_1x1;
		line_update_mode_2 = line_update_mode2_1x1;
		line_update_mode_3 = line_update_mode3_1x1;
		line_update_mode_4 = line_update_mode4_1x1;
	}

  scalefactor=value;

	/* Make sure we update the whole screen */
  memset(LineNeedsUpdating,1,sizeof(LineNeedsUpdating));
	BorderChangeInLine=totallines-1;
	
  //must be immediately set by call to VideoSetMode()
  line_update=0;
}


//---------------------------------------------------------------------------


int VideoSwitchAutoScale(void) {
//
//turn on/off autoscale
//
  return cfg.autoscale^=1;
}


//---------------------------------------------------------------------------


void VideoSetMode(int mode) {
//
//set Sam video mode to emulate
//
  if(mode==3) {
    //autoscale to 2x2
    if(cfg.autoscale && scalefactor==1) VideoSetScale(2);
    //if 2x2 scale not available, turn off autoscale
    if(scalefactor!=2) cfg.autoscale=0;
    line_update=line_update_mode_3;
    mode3begin=2;

  } else {

    //autoscale to 1x1
    if(cfg.autoscale && scalefactor==2) rescaledown=mode;

    switch(mode) {
    case 1: line_update = line_update_mode_1; break;
    case 2: line_update = line_update_mode_2; break;
    case 4: line_update = line_update_mode_4; break;
    }
  }
}


//---------------------------------------------------------------------------


void VideoScreenShot(const char *fname,int border) {
//
//save screenshot in current video resolution
//
//entry: fname  = filename (with or without extension)
//       border = border flag
//
  BITMAP *bmp;
  PALETTE pal;

  int w=SCREEN_W;
  int h=SCREEN_H;

  if(!border) {
    w&=~255;
    h-=(h%192);
  }

  get_palette(pal);
  bmp=create_sub_bitmap(screen,(SCREEN_W-w)/2,(SCREEN_H-h)/2,w,h);
  save_pcx((char*)fname,bmp,pal);
  destroy_bitmap(bmp);
}

